var dmtPreferencesService = null;
var dmtPrevSelectedCount = 0;
var dmtPrevSelectedItem = null;
var dmtDownloadManager = null;
var dmtDownloadListener = null;
var dmtObserverService = null;
var dmtSavedDownloadsView = null;
var dmtProjectName = "Download Manager Tweak";
var dmtReloadButtonLabel = "";
var dmtLaunchButtonLabel = "";
var dmtShowButtonLabel = "";
var dmtDeleteFileButtonLabel = "";
var dmtCleanUpButtonLabel = "";
var dmtOptionsButtonLabel = "";
var dmtDeleteConfirmRadio = 0;
var dmtButtonClass = " DMT_miniButton DMT_miniButton_big";
var dmtIconClass = " DMT_miniIcon DMT_miniIcon_big";
var dmtThisIsASidebar = false;
var dmtStartupCompleted = false;
var dmtShutdownInStartup = false;

////////////////////////////////////////////////////////////////////////////////
// This event listener is called whenever the selection changes for the
// download list. Need to selectively enable/disable the command bar buttons
// and fill in the infoBar data.
function dmtOnDownloadSelect(aEvent)
{
  var dlState = 0;

  // Make sure selection changed so we don't redraw too often.
  if ((gDownloadsView.selectedCount != dmtPrevSelectedCount) ||
      (gDownloadsView.selectedItem != dmtPrevSelectedItem))
  {
    dmtPrevSelectedCount = gDownloadsView.selectedCount;
    dmtPrevSelectedItem = gDownloadsView.selectedItem;

    // Get parameter for dmtUpdateToolInfoBar().
    if ((gDownloadsView.selectedCount > 0) && gDownloadsView.selectedItem)
    {
      dlState = parseInt(gDownloadsView.selectedItem.getAttribute("state"));
    }
    dmtUpdateToolInfoBar(dlState);
    dmtUpdateDownloadButtons(gDownloadsView.selectedItem);
  }
}

////////////////////////////////////////////////////////////////////////////////
// Updates all the toolbar and infobar parts which change when different
// downloads are selected or the selected changes state.
function dmtUpdateToolInfoBar(dlState)
{
  // Create variables, initialized to case of no download selected.
  var reloadDisable     = true;
  var launchDisable     = true;
  var showDisable       = true;
  var deleteFileDisable = true;
  var cleanUpDisable    = true;
  var from = "";
  var to   = "";
  var date = "";

  if ((gDownloadsView.selectedCount > 0) && gDownloadsView.selectedItem)
  {
    var fileExists = false;
    var dirExists = false;
    var dateObj = new Date(parseInt(gDownloadsView.selectedItem.getAttribute("endTime")));

    // Fill in the infoBar data according to the selected download.
    from = gDownloadsView.selectedItem.getAttribute("uri");
    to   = gDownloadsView.selectedItem.getAttribute("path");
    date = dateObj.toLocaleString();

    // Enable buttons for the selected download according to the state of
    // that download.
    switch (dlState) {
    case nsIDM.DOWNLOAD_FINISHED:
      // Check to see if the file for selected download actually exists. It may
      // have been deleted or the download didn't complete.
      try {
        var fileName = gDownloadsView.selectedItem.getAttribute("file");
        var fileObj = getLocalFileFromNativePathOrUrl(fileName);
        fileExists = fileObj.exists();
        dirExists = fileObj.parent.exists();
      } catch (e) {
        fileExists = false;
        dirExists = false;
      }

      // Only allow some buttons if a file exists. This same criteria is used
      // by the command handlers so disabling these doesn't really prevent the
      // command.
      if (fileExists)
      {
        launchDisable = false;
        deleteFileDisable = false;
      }
      else
        reloadDisable = false;
      if (dirExists)
        showDisable = false;
      cleanUpDisable = false;
      break;
    case nsIDM.DOWNLOAD_FAILED:
    case nsIDM.DOWNLOAD_CANCELED:
      reloadDisable = false;
      cleanUpDisable = false;
      break;
    case nsIDM.DOWNLOAD_PAUSED:
    case nsIDM.DOWNLOAD_NOTSTARTED:
    case nsIDM.DOWNLOAD_DOWNLOADING:
    case nsIDM.DOWNLOAD_QUEUED:
    case nsIDM.DOWNLOAD_BLOCKED_PARENTAL:
    case nsIDM.DOWNLOAD_SCANNING:
    case nsIDM.DOWNLOAD_DIRTY:
    case nsIDM.DOWNLOAD_BLOCKED_POLICY:
    default:
      // All buttons remain disabled.
      break;
    }
  }

  // Display either reload or launch button and not both.
  var hideLaunchButton = !dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.buttonLaunch");
  if (!hideLaunchButton)
  {
    document.getElementById("dmtLaunchButton").hidden      = !reloadDisable;
    document.getElementById("dmtLaunchButton2").hidden     = !reloadDisable;
  }
  var hideReloadButton = !dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.buttonDownload");
  if (!hideReloadButton)
  {
    document.getElementById("dmtReloadButton").hidden      = reloadDisable;
    document.getElementById("dmtReloadButton2").hidden     = reloadDisable;
  }

  // Set the button disables as calculated.
  document.getElementById("dmtReloadButton").disabled      = reloadDisable;
  document.getElementById("dmtReloadButton2").disabled     = reloadDisable;
  document.getElementById("dmtLaunchButton").disabled      = launchDisable;
  document.getElementById("dmtLaunchButton2").disabled     = launchDisable;
  document.getElementById("dmtShowButton").disabled        = showDisable;
  document.getElementById("dmtShowButton2").disabled       = showDisable;
  document.getElementById("dmtDeleteFileButton").disabled  = deleteFileDisable;
  document.getElementById("dmtDeleteFileButton2").disabled = deleteFileDisable;
  document.getElementById("dmtCleanUpButton").disabled     = cleanUpDisable;
  document.getElementById("dmtCleanUpButton2").disabled    = cleanUpDisable;

  // Fill in the infoBar with generated data.
  document.getElementById("dmtDownloadFrom").value  = from;
  document.getElementById("dmtDownloadFrom2").value = from;
  document.getElementById("dmtDownloadTo").value    = to;
  document.getElementById("dmtDownloadTo2").value   = to;
  document.getElementById("dmtDownloadDate").value  = date;
  document.getElementById("dmtDownloadDate2").value = date;
}

////////////////////////////////////////////////////////////////////////////////
// DMT changes the icons according to user selected options. The .css files
// use the button class to change the button appearance. The updates the classes
// for all the buttons in the download.
function dmtUpdateDownloadIconClasses(thisDownload)
{
  try {
    // Apply the button class to all the buttons.
    for (i = 0; i < thisDownload.buttons.length; i++)
    {
      var classStrings = String(thisDownload.buttons[i].className).split(" ");

      if ((classStrings[1] == "mini-button") ||
          (classStrings[1] == "DMT_miniButton") ||
          (classStrings[1] == "DMT_miniButton_off"))
      {
        // This is an original button.
//        var newString = classStrings[0] + " mini-button" + dmtButtonClass;
        var newString = classStrings[0] + dmtButtonClass;

        if (thisDownload.buttons[i].className != newString)
          thisDownload.buttons[i].className = newString
      }
      else if ((classStrings[1] == "DMT_miniIcon") ||
               (classStrings[1] == "DMT_miniIcon_off"))
      {
        // This is a DMT added button.
        var newString = classStrings[0] + dmtIconClass;

        if (thisDownload.buttons[i].className != newString)
          thisDownload.buttons[i].className = newString
      }
    }
  } catch (e) {}
}

////////////////////////////////////////////////////////////////////////////////
// DMT adds new buttons to the downloads. For the download finished case, some
// of these buttons don't apply so those need to be hidden. Note they are hidden
// and not disabled because updateButtons() in the download manager changes the
// disables by it's own criteria, not DMT's.
function dmtUpdateDownloadButtons(thisDownload)
{
  dmtUpdateDownloadIconClasses(thisDownload);
  try {
    var dlState = parseInt(thisDownload.getAttribute("state"));

    var miniButtonType = dmtPreferencesService.getIntPref(
                                        "extensions.downloadmgr.miniButtonType");

    if (dmtThisIsASidebar)
    {
      // For sidebars force button type to icon-only, as the other types won't fit.
      miniButtonType = 1;
    }

    switch (dlState)
    {
    case nsIDM.DOWNLOAD_FINISHED:
      {
        var showButton       = null;
        var reloadButton     = null;
        var deleteFileButton = null;
        var cleanUpButton    = null;
        var hideShowMiniButton = !dmtPreferencesService.getBoolPref(
                                             "extensions.downloadmgr.miniButtonShow");
        var hideReloadMiniButton = false;
        var hideReloadMiniButton = !dmtPreferencesService.getBoolPref(
                                             "extensions.downloadmgr.miniButtonDownload");
        var hideDeleteMiniButton = !dmtPreferencesService.getBoolPref(
                                             "extensions.downloadmgr.miniButtonDelete");
        var hideCleanMiniButton = !dmtPreferencesService.getBoolPref(
                                             "extensions.downloadmgr.miniButtonClean");

        // Verify that this is the correct download display type with the DMT
        // added buttons. It could be that display hasn't updated to match the
        // download state or that DMT theme is disabled and buttons will never
        // be there.
        if (thisDownload.buttons.length >= 4)
        {
          showButton       = thisDownload.buttons[0];
          reloadButton     = thisDownload.buttons[1];
          deleteFileButton = thisDownload.buttons[2];
          cleanUpButton    = thisDownload.buttons[3];

          if (showButton.getAttribute("id") != "dmtDoneShowBtn")
            showButton = null;
          if (reloadButton.getAttribute("id") != "dmtDoneReloadBtn")
            reloadButton = null;
          if (deleteFileButton.getAttribute("id") != "dmtDoneDeleteBtn")
            deleteFileButton = null;
          if (cleanUpButton.getAttribute("id") != "dmtDoneCleanUpBtn")
            cleanUpButton = null;
        }

        // If any of these are null then this isn't the correct download type.
        if (reloadButton && deleteFileButton && showButton && cleanUpButton)
        {
          var showDisable       = hideShowMiniButton;
          var reloadDisable     = hideReloadMiniButton;
          var deleteFileDisable = hideDeleteMiniButton;

          // Always hide if user disabled them.
          if (!hideShowMiniButton || !hideReloadMiniButton || !hideDeleteMiniButton)
          {
            // Check to see if the file for this download actually exists. It may
            // have been deleted.
            try {
              var fileName   = thisDownload.getAttribute("file");
              var fileObj    = getLocalFileFromNativePathOrUrl(fileName);
              var fileExists = fileObj.exists();
              var dirExists  = fileObj.parent.exists();

              if (!dirExists)
                showDisable = true;
              if (fileExists)
                reloadDisable = true;
              else
                deleteFileDisable = true;
            } catch (e) {
              // Disables will all be true so hide every button.
            }
          }

          if (miniButtonType == 1)
          {
            showButton.removeAttribute("label");
            reloadButton.removeAttribute("label");
            deleteFileButton.removeAttribute("label");
            cleanUpButton.removeAttribute("label");
          }
          else
          {
            showButton.label       = dmtShowButtonLabel;
            reloadButton.label     = dmtReloadButtonLabel;
            deleteFileButton.label = dmtDeleteFileButtonLabel;
            cleanUpButton.label    = dmtCleanUpButtonLabel;
          }

          // Hide the DMT added buttons as determined.
          showButton.hidden       = showDisable;
          reloadButton.hidden     = reloadDisable;
          deleteFileButton.hidden = deleteFileDisable;
          cleanUpButton.hidden    = hideCleanMiniButton;
        }
        else
          return false;
      }
      break;

    case nsIDM.DOWNLOAD_BLOCKED_PARENTAL:
    case nsIDM.DOWNLOAD_SCANNING:
    case nsIDM.DOWNLOAD_DIRTY:
    case nsIDM.DOWNLOAD_BLOCKED_POLICY:
      // No buttons displayed in these states.
      break;

    case nsIDM.DOWNLOAD_FAILED:
    case nsIDM.DOWNLOAD_CANCELED:
      {
        var hideCleanMiniButton = !dmtPreferencesService.getBoolPref(
                                             "extensions.downloadmgr.miniButtonClean");
        var retryButton = null;
        var cleanUpButton = null;

        if (thisDownload.buttons.length >= 2)
        {
          if ((thisDownload.buttons[1].getAttribute("id") == "dmtCancelCleanUpBtn") ||
              (thisDownload.buttons[1].getAttribute("id") == "dmtFailCleanUpBtn"))
          {
            retryButton = thisDownload.buttons[0];
            cleanUpButton = thisDownload.buttons[1];
          }
        }

        if (retryButton && cleanUpButton)
        {
          if (miniButtonType == 1)
          {
            retryButton.removeAttribute("label");
            cleanUpButton.removeAttribute("label");
          }
          else
          {
            retryButton.label = retryButton.tooltipText;
            cleanUpButton.label = dmtCleanUpButtonLabel;
          }

          // Hide the DMT added button as determined.
          cleanUpButton.hidden = hideCleanMiniButton;
        }
        else
          return false;
      }
      break;

    case nsIDM.DOWNLOAD_DOWNLOADING:
      {
        var pauseButton = null;
        var cancelButton = null;

        if (thisDownload.buttons.length >= 2)
        {
          //if ((thisDownload.buttons[1].getAttribute("id") == "dmtCancelCleanUpBtn") ||
          //    (thisDownload.buttons[1].getAttribute("id") == "dmtFailCleanUpBtn"))
          {
            pauseButton = thisDownload.buttons[0];
            cancelButton = thisDownload.buttons[1];
          }
        }

        if (pauseButton && cancelButton)
        {
          if (miniButtonType == 1)
          {
            pauseButton.removeAttribute("label");
            cancelButton.removeAttribute("label");
          }
          else
          {
            pauseButton.label = pauseButton.tooltipText;
            cancelButton.label = cancelButton.tooltipText;
          }
        }
        else
          return false;
      }
      break;

    case nsIDM.DOWNLOAD_PAUSED:
      {
        var resumeButton = null;
        var cancelButton = null;

        if (thisDownload.buttons.length >= 2)
        {
          //if ((thisDownload.buttons[1].getAttribute("id") == "dmtCancelCleanUpBtn") ||
          //    (thisDownload.buttons[1].getAttribute("id") == "dmtFailCleanUpBtn"))
          {
            resumeButton = thisDownload.buttons[0];
            cancelButton = thisDownload.buttons[1];
          }
        }

        if (resumeButton && cancelButton)
        {
          if (miniButtonType == 1)
          {
            resumeButton.removeAttribute("label");
            cancelButton.removeAttribute("label");
          }
          else
          {
            resumeButton.label = resumeButton.tooltipText;
            cancelButton.label = cancelButton.tooltipText;
          }
        }
        else
          return false;
      }
      break;

    case nsIDM.DOWNLOAD_NOTSTARTED:
    case nsIDM.DOWNLOAD_QUEUED:
    default:
      {
        var cancelButton = null;

        if (thisDownload.buttons.length >= 1)
        {
          //if ((thisDownload.buttons[1].getAttribute("id") == "dmtCancelCleanUpBtn") ||
          //    (thisDownload.buttons[1].getAttribute("id") == "dmtFailCleanUpBtn"))
          {
            cancelButton = thisDownload.buttons[0];
            if (miniButtonType == 1)
              cancelButton.removeAttribute("label");
            else
              cancelButton.label = cancelButton.tooltipText;
          }
        }
        else
          return false;
      }
      break;
    }

    // Apply the button classes to all the buttons.
    dmtUpdateDownloadIconClasses(thisDownload);
  } catch (e) {}
  return true;
}

////////////////////////////////////////////////////////////////////////////////
// DMT adds new buttons to the downloads. For the download finished case, some
// of these buttons don't apply so those need to be hidden. This loops through
// the entire list of downloads and updates the buttons in each one.
function dmtUpdateDownloadViewButtons()
{
  for (let i = 0; i < gDownloadsView.itemCount; i++)
  {
    dmtUpdateDownloadButtons(gDownloadsView.getItemAtIndex(i));
  }
}

////////////////////////////////////////////////////////////////////////////////
// gDownloadsView changes on occasion, for example whenever buildDownloadList()
// is called. That results in the event listener no longer generating select
// events. Need to remove the listener for the old view and and add one for the
// new view. This is only done if gDownloadsView changes.
function dmtOnDownloadClick(aEvent)
{
  if (dmtSavedDownloadsView != gDownloadsView)
  {
    dmtSavedDownloadsView.removeEventListener("select", dmtOnDownloadSelect,
                                              false);
    gDownloadsView.addEventListener("select", dmtOnDownloadSelect, false);
    dmtSavedDownloadsView = gDownloadsView;
    dmtOnDownloadSelect(aEvent);
  }
}

////////////////////////////////////////////////////////////////////////////////
// Called whenever the DMT window/tab gets focus. This updates the status of the
// buttons in case user has deleted a file outside Firefox. Deleting a file
// won't change the buttons instantly, but they will change as soon as the
// window is selected.
function dmtOnDownloadFocus(aEvent)
{
  var dlState = 0;

  // Get parameter for dmtUpdateToolInfoBar().
  if ((gDownloadsView.selectedCount > 0) && gDownloadsView.selectedItem)
  {
    dlState = parseInt(gDownloadsView.selectedItem.getAttribute("state"));
  }

  // Update status of buttons in the toolbar and the download view.
  dmtUpdateToolInfoBar(dlState);
  dmtUpdateDownloadViewButtons();
}

////////////////////////////////////////////////////////////////////////////////
// Hides and shows window elements depending on the preferences set by the user.
// Can be called at startup or whenever the preferences are changed.
function dmtHideByPreference()
{
  var unhideList = ["dmtReloadButton",     "dmtReloadButton2",
                    "dmtLaunchButton",     "dmtLaunchButton2",
                    "dmtShowButton",       "dmtShowButton2",
                    "dmtDeleteFileButton", "dmtDeleteFileButton2",
                    "dmtCleanUpButton",    "dmtCleanUpButton2",
                    "dmtOptionsButton",    "dmtOptionsButton2",
                    "dmtTopToolBox",       "dmtBottomToolBox",
                    "dmtTopToolBar",       "dmtBottomToolBar",
                    "dmtDownloadInfoBar",  "dmtDownloadInfoBar2"];
  var actionButtons = ["dmtReloadButton",     "dmtReloadButton2",
                       "dmtLaunchButton",     "dmtLaunchButton2",
                       "dmtShowButton",       "dmtShowButton2",
                       "dmtDeleteFileButton", "dmtDeleteFileButton2",
                       "dmtCleanUpButton",    "dmtCleanUpButton2",
                       "dmtOptionsButton",    "dmtOptionsButton2"];

  try {
    // First unhide everything to match original setting in .xul.
    for (i = 0; i < unhideList.length; i++)
      document.getElementById(unhideList[i]).hidden = false;
  } catch (e) {}

  try {
    var showOnBottom = dmtPreferencesService.getBoolPref(
                                        "extensions.downloadmgr.bottomToolBar");
    if (showOnBottom)
      document.getElementById("dmtTopToolBox").hidden = true;
    else
      document.getElementById("dmtBottomToolBox").hidden = true;

    var hideLaunchButton = !dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.buttonLaunch");
    if (hideLaunchButton) {
      document.getElementById("dmtLaunchButton").hidden  = true;
      document.getElementById("dmtLaunchButton2").hidden = true;
    }

    var hideReloadButton = !dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.buttonDownload");
    // Always hide these now, maybe unhidden later?
    document.getElementById("dmtReloadButton").hidden  = true;
    document.getElementById("dmtReloadButton2").hidden = true;

    var hideShowButton = !dmtPreferencesService.getBoolPref(
                                           "extensions.downloadmgr.buttonShow");
    if (hideShowButton) {
      document.getElementById("dmtShowButton").hidden  = true;
      document.getElementById("dmtShowButton2").hidden = true;
    }

    var hideDeleteFileButton = !dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.buttonDelete");
    if (hideDeleteFileButton) {
      document.getElementById("dmtDeleteFileButton").hidden  = true;
      document.getElementById("dmtDeleteFileButton2").hidden = true;
    }

    var hideCleanUpButton = !dmtPreferencesService.getBoolPref(
                                          "extensions.downloadmgr.buttonClean");
    if (hideCleanUpButton) {
      document.getElementById("dmtCleanUpButton").hidden  = true;
      document.getElementById("dmtCleanUpButton2").hidden = true;
    }

    var hideOptionsButton = !dmtPreferencesService.getBoolPref(
                                        "extensions.downloadmgr.buttonOptions");
    if (hideOptionsButton) {
      document.getElementById("dmtOptionsButton").hidden  = true;
      document.getElementById("dmtOptionsButton2").hidden = true;
    }

    // If all buttons hidden then hide the entire toolbar.
    if (hideLaunchButton  && hideReloadButton && hideShowButton &&
        hideDeleteFileButton && hideCleanUpButton && hideOptionsButton)
    {
      document.getElementById("dmtTopToolBar").hidden  = true;
      document.getElementById("dmtBottomToolBar").hidden = true;
    }

    // Display all buttons as: 0 - icon and text, 1 - icon, 2 - text
    var toolButtonType = dmtPreferencesService.getIntPref(
                                        "extensions.downloadmgr.toolButtonType");
    var miniButtonType = dmtPreferencesService.getIntPref(
                                        "extensions.downloadmgr.miniButtonType");
    if (dmtThisIsASidebar)
    {
      // For sidebars force button type to icon-only, as the other types won't fit.
      toolButtonType = 1;
      miniButtonType = 1;
    }
    if (toolButtonType == 1)
    {
      for (i = 0; i < actionButtons.length; i++)
        document.getElementById(actionButtons[i]).removeAttribute("label");
    }
    else
    {
      document.getElementById("dmtReloadButton").label      = dmtReloadButtonLabel;
      document.getElementById("dmtReloadButton2").label     = dmtReloadButtonLabel;
      document.getElementById("dmtLaunchButton").label      = dmtLaunchButtonLabel;
      document.getElementById("dmtLaunchButton2").label     = dmtLaunchButtonLabel;
      document.getElementById("dmtShowButton").label        = dmtShowButtonLabel;
      document.getElementById("dmtShowButton2").label       = dmtShowButtonLabel;
      document.getElementById("dmtDeleteFileButton").label  = dmtDeleteFileButtonLabel;
      document.getElementById("dmtDeleteFileButton2").label = dmtDeleteFileButtonLabel;
      document.getElementById("dmtCleanUpButton").label     = dmtCleanUpButtonLabel;
      document.getElementById("dmtCleanUpButton2").label    = dmtCleanUpButtonLabel;
      document.getElementById("dmtOptionsButton").label     = dmtOptionsButtonLabel;
      document.getElementById("dmtOptionsButton2").label    = dmtOptionsButtonLabel;
    }

    // Figure out the button class to use according to user preferences.
    var toolButtonSize = dmtPreferencesService.getIntPref(
                                        "extensions.downloadmgr.toolButtonSize");
    var miniButtonSize = dmtPreferencesService.getIntPref(
                                        "extensions.downloadmgr.miniButtonSize");
    var toolButtonClass = "";
    if (toolButtonSize == 0)
    {
      toolButtonClass = "DMT_icon DMT_icon_big";
    }
    else
    {
      toolButtonClass = "DMT_icon DMT_icon_small";
    }

    // Apply the button class to all the toolbar buttons.
    if (toolButtonType == 2)
    {
      for (i = 0; i < actionButtons.length; i++)
        document.getElementById(actionButtons[i]).className = "DMT_icon_off";
    }
    else
    {
      for (i = 0; i < actionButtons.length; i++)
        document.getElementById(actionButtons[i]).className = toolButtonClass;
    }

    // Figure out the mini-button class to use according to user preferences.
    if (miniButtonType == 2)
    {
        dmtButtonClass = " DMT_miniButton_off";
        dmtIconClass = " DMT_miniIcon_off";
    }
    else
    {
      if (miniButtonSize == 0)
      {
        dmtButtonClass = " DMT_miniButton DMT_miniButton_big";
        dmtIconClass = " DMT_miniIcon DMT_miniIcon_big";
      }
      else
      {
        dmtButtonClass = " DMT_miniButton DMT_miniButton_small";
        dmtIconClass = " DMT_miniIcon DMT_miniIcon_small";
      }
    }

    var hideDmtTheme = dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.hideDmtTheme");
    if (hideDmtTheme)
      dmtButtonClass = " mini-button";

    var hideInfoBar = !dmtPreferencesService.getBoolPref(
                                          "extensions.downloadmgr.showInfoBar");
    if (hideInfoBar) {
      document.getElementById("dmtDownloadInfoBar").hidden  = true;
      document.getElementById("dmtDownloadInfoBar2").hidden = true;
    }
  } catch (e) {alert("DMT dmtHideByPreference() failed:\n" + e.name + "\n" + e.message);}
}

////////////////////////////////////////////////////////////////////////////////
// This observer is called whenever preference in the "extensions.downloadmgr."
// branch changes.
var dmtPrefObserver = {
  observe: function (aSubject, aTopic, aData)
  {
    try {
      if (aTopic == "nsPref:changed")
      {
        if (aData == "extensions.downloadmgr.windowType")
        {
          // Need to close window with old type and then open with new type.
          // Unfortunately I don't think this can be done here.
          //alert("DMT observe: window type changed");
        }
        else
        {
          // Just redraw existing window.
          dmtHideByPreference();
          dmtUpdateDownloadViewButtons();
        }
      }
    } catch (e) {}
  }
};

////////////////////////////////////////////////////////////////////////////////
// Setup the extended features added by DMT and then startup the original
// download manger being tweaked.
function dmtStartup()
{
  dmtStartupCompleted = false;
  dmtShutdownInStartup = false;
  dmtPreferencesService = Components.classes["@mozilla.org/preferences-service;1"].
                          getService(Components.interfaces.nsIPrefService).getBranch("");
  dmtDownloadManager = document.getElementById("downloadManager");
  dmtObserverService = Components.classes["@mozilla.org/observer-service;1"].
                       getService(Components.interfaces.nsIObserverService);

  // Get some values that are available in the .XUL file, but not easily
  // available here.
  try {
    dmtProjectName           = document.getElementById("dmtDummyLabel").value;
    dmtReloadButtonLabel     = document.getElementById("dmtReloadButton").label;
    dmtLaunchButtonLabel     = document.getElementById("dmtLaunchButton").label;
    dmtShowButtonLabel       = document.getElementById("dmtShowButton").label;
    dmtDeleteFileButtonLabel = document.getElementById("dmtDeleteFileButton").label;
    dmtCleanUpButtonLabel    = document.getElementById("dmtCleanUpButton").label;
    dmtOptionsButtonLabel    = document.getElementById("dmtOptionsButton").label;

    // The escape key was not assigned an ID inside the XUL so it can't be
    // overridden like all the other keys. This does it the hard way.
    var actionKeys = document.getElementsByTagName("key");
    for (var j = 0; j < actionKeys.length; j++)
    {
      if (actionKeys[j].getAttribute("keycode") == "VK_ESCAPE")
      {
        actionKeys[j].setAttribute("oncommand", "dmtEscapeDownloadMgr();");
      }
    }

    var hideDmtTheme = dmtPreferencesService.getBoolPref(
                                         "extensions.downloadmgr.hideDmtTheme");
    // User wants to use their own theme. Disable the CCS rules for the DMT
    // "theme". These are limited to changes of the underlying download view
    // and not the new to DMT CSS rules.
    for (var i = 0; i < document.styleSheets.length; i++)
    {
      //if (document.styleSheets[i].href ==
      //    'chrome://downloadmgr/skin/dmtAddedMiniButtons.css')
      //{
      //  if (hideDmtTheme)
      //  {
      //    // Found the rules for the DMT theme, delete them all.
      //    for (var j = document.styleSheets[i].cssRules.length - 1; j >= 0; j--)
      //    {
      //      document.styleSheets[i].deleteRule(0);
      //    }
      //  }
      //}
      if (document.styleSheets[i].href ==
          'chrome://downloadmgr/skin/dmtThemeSkin.css')
      {
        //if (hideDmtTheme)
        //  document.styleSheets[i].disabled = true;
        //else
        //  document.styleSheets[i].disabled = false;

        // Need to reload the document for the above to work. Couldn't figure
        // out how so use the below.
        if (hideDmtTheme)
        {
          // Found the rules for the DMT theme, delete them all.
          for (var j = document.styleSheets[i].cssRules.length - 1; j >= 0; j--)
          {
            document.styleSheets[i].deleteRule(0);
          }
        }
      }
//      else if (document.styleSheets[i].href ==
////               'chrome://mozapps/skin/downloads/downloads.css')
//               'chrome://global/skin/button.css')
//      {
////        var re = new RegExp("#downloadManager button,");
//
//        // Found the rules for the DMT theme, delete them all.
//        for (var j = document.styleSheets[i].cssRules.length - 1; j >= 0; j--)
//        {
//          var ruleSelector = String(document.styleSheets[i].cssRules[j].selectorText);
//
////          if (ruleSelector.search(re) != -1)
//          if (ruleSelector.indexOf("#downloadManager button,") != -1)
//
//          {
////            alert(ruleSelector);
////            var rs1 = ruleSelector.replace("#downloadManager button,", "");
////            alert(rs1);
////            var rs2 = rs1.replace("#downloadManager button > .button-box,", "");
////            alert(rs2);
//            document.styleSheets[i].deleteRule(j);
//          }
//        }
//      }
    }

    // Check to see if this window is a sidebar. All sidebars have the same
    // fairly unique class that we can search on. Can't use the element ID
    // because multiple sidebars have different IDs.
    var sidebarElements = window.top.document.getElementsByClassName(
                                                  "chromeclass-extrachrome");
    if (sidebarElements.length > 0)
    {
      dmtThisIsASidebar = true;
    }
  } catch (e) {alert("DMT Startup() failed:\n" + e.name + "\n" + e.message);}

  dmtHideByPreference();

  // Startup() is inside mozapps/content/downloads/downloads.js, the download
  // manager built into FireFox. It requires that "window.arguments[1]" be
  // defined or it will crash. arguments[0] is the download that triggered the
  // download manager to open and is ignored by downloads.js. arguments[1] is
  // the reason the download manager is being opened (by download starting or
  // user selecting to open the window). The selected value causes downloads.js
  // to not autoclose the window, DMT does that.
  try {
    window.arguments = [0, Ci.nsIDownloadManagerUI.REASON_USER_INTERACTED];
    Startup();
    delete window.arguments;
  } catch (e) {}//alert("DM Startup() failed:\n" + e.name + "\n" + e.message);}

//  try {
//    // Hopefully this will hide the menubar on OSX.
//    window.menubar.visible=false;
//    window.menubar.hidden=true;
//  } catch (e) {}

  // Can not do this here because the download view is not populated with
  // downloads yet. An observer setup below will handle this.
  //dmtUpdateDownloadViewButtons();

  // Setup listeners so that the command/info bar can be updated.
  try {
    // Need to change the button enables and infoBar data as different
    // downloads are selected.
    gDownloadsView.addEventListener("select", dmtOnDownloadSelect, false);
    dmtSavedDownloadsView = gDownloadsView;
    dmtDownloadManager.addEventListener("click", dmtOnDownloadClick, false);

    // This listens for window focus events, so that the buttons can be updated
    // in case user has deleted a file outside Firefox.
    addEventListener("focus", dmtOnDownloadFocus, false);

    // This listens for download state changes and updates the button enables.
    // For example, you can't launch a file that is still downloading, but you
    // can once the download is complete.
    dmtDownloadListener = new dmtDownloadProgressListener();
    gDownloadManager.addListener(dmtDownloadListener);

    // Listens for changes to the preferences (mainly on the options popup)
    // and immediately redraws the screen to reflect the changes.
    dmtPreferencesService.QueryInterface(Components.interfaces.nsIPrefBranch2);
    dmtPreferencesService.addObserver("extensions.downloadmgr.",
                                      dmtPrefObserver, false);

    // This handles the events generated when downloads have been removed from
    // the list. May need to update the tool and info bar if the selected
    // download has been removed.
    dmtObserverService.addObserver(dmtDownloadmgrOverlayObserver,
                                   "download-manager-remove-download", false);
    // This handles the events generated when the download view has finished
    // being filled with downloads. That is when it is time to call
    // dmtUpdateDownloadViewButtons().
    dmtObserverService.addObserver(dmtDownloadmgrOverlayObserver,
                                   "download-manager-ui-done", false);
  } catch (e) {}//alert("DMT Startup() failed:\n" + e.name + "\n" + e.message);}

  dmtStartupCompleted = true;
  if (dmtShutdownInStartup)
    dmtShutdown()
}

////////////////////////////////////////////////////////////////////////////////
// Cleanup the extended features added by DMT and then shutdown the original
// download manger being tweaked.
function dmtShutdown()
{
  if (dmtStartupCompleted)
  {
    //Services.console.logStringMessage("leanup in dmtShutdown()");
    dmtStartupCompleted = false;
    dmtShutdownInStartup = false;
    // Cleanup listeners.
    try {
      dmtSavedDownloadsView.removeEventListener("select", dmtOnDownloadSelect,
                                                false);
      gDownloadsView.removeEventListener("select", dmtOnDownloadSelect, false);
      dmtDownloadManager.removeEventListener("click", dmtOnDownloadClick, false);

      gDownloadManager.removeListener(dmtDownloadListener);

      dmtPreferencesService.removeObserver("extensions.downloadmgr.",
                                           dmtPrefObserver);
      dmtPreferencesService.QueryInterface(Components.interfaces.nsIPrefBranch2);

      dmtObserverService.removeObserver(dmtDownloadmgrOverlayObserver,
                                        "download-manager-remove-download");
      dmtObserverService.removeObserver(dmtDownloadmgrOverlayObserver,
                                        "download-manager-ui-done");
    } catch (e) {}//alert("DMT Shutdown() failed:\n" + e.name + "\n" + e.message);}

    // Shutdown() is inside mozapps/content/downloads/downloads.js, the download
    // manager built into FireFox.
    try {
      Shutdown();
    } catch (e) {}//alert("DM Shutdown() failed:\n" + e.name + "\n" + e.message);}
  }
  else
  {
    // If dmtShutdown is happening before dmtStartup has had a chance to
    // complete, then reschedule dmtShutdown to happen after dmtStartup has
    // had time to complete.
    dmtShutdownInStartup = true;
  }
}

////////////////////////////////////////////////////////////////////////////////
// Called when the options button is pressed. Displays the options dialog.
function dmtShowOptions()
{
  var features;
  try {
    var instantApply = dmtPreferencesService.getBoolPref("browser.preferences.instantApply");
    features = "chrome,titlebar,toolbar,centerscreen" + (instantApply ? ",dialog=no" : ",modal");
  }
  catch (e) {
    features = "chrome,titlebar,toolbar,centerscreen,modal";
  }
  openDialog("chrome://downloadmgr/content/options/dmtOptions.xul", "", features);
}

////////////////////////////////////////////////////////////////////////////////
// Loops through all the selected downloads and reloads the associated file.
// Only done if file doesn't exists. Uses retry if load failed or was canceled
// and does something special if download finished but file was deleted.
function dmtReloadFiles(aItem)
{
  var reloadlist = [];

  if (aItem)
  {
    // Reload the specified item, first need to get the download.
    var aDownload = aItem;
    while (aDownload.nodeName != "richlistitem" ||
           aDownload.getAttribute("type") != "download")
      aDownload = aDownload.parentNode;

    // Put this download in the list by itself.
    reloadlist[0] = aDownload;
  }
  else
  {
    // Reload all the files in the list. Need to copy the list because the
    // reload process changes the original list.
    reloadlist[gDownloadsView.selectedItems.length - 1] = null;
    for (let i = 0; i < gDownloadsView.selectedItems.length; i++)
      reloadlist[i] = gDownloadsView.selectedItems[i];
  }

  // Do a file reload for each download in the list.
  for each (let thisDownload in reloadlist)
  {
    try
    {
      // Get a file object from the file name.
      var fileName = thisDownload.getAttribute("file");
      //var filePath = thisDownload.getAttribute("path");
      var fileObj = getLocalFileFromNativePathOrUrl(fileName);
      var dlState = parseInt(thisDownload.getAttribute("state"));

      // Make sure file doesn't exist before trying to redownload it.
      if (!fileObj.exists())
      {
        switch (dlState) {
        case nsIDM.DOWNLOAD_FAILED:
        case nsIDM.DOWNLOAD_CANCELED:
          // Use the standard retry if possible.
          gDownloadViewController.doCommand('cmd_retry', thisDownload);
          break;

        case nsIDM.DOWNLOAD_FINISHED:
          // Can't use retry, have to create a new download.
          const nsIWBP = Ci.nsIWebBrowserPersist;
          var ios = Cc["@mozilla.org/network/io-service;1"].
                    getService(Ci.nsIIOService);
          var fromURI = ios.newURI(thisDownload.getAttribute('uri'), null, null);
          var toURI = ios.newFileURI(fileObj);
          var persist = null;
          var dl = null;

          // Remove the download from the list, a new one will be created below.
          gDownloadViewController.doCommand('cmd_removeFromList', thisDownload);

          // Create a persist object. This controls and monitors the download.
          persist = Cc["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                    .createInstance(Ci.nsIWebBrowserPersist);
          persist.persistFlags = nsIWBP.PERSIST_FLAGS_REPLACE_EXISTING_FILES |
                                 nsIWBP.PERSIST_FLAGS_BYPASS_CACHE |
                                 nsIWBP.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;

          // Create a new download.
          dl = gDownloadManager.addDownload(
                            Ci.nsIDownloadManager.DOWNLOAD_TYPE_DOWNLOAD,
                            fromURI, toURI, null, null, 0, null, persist);

          // Set the progress listener to the download object and then start
          // the download.
          persist.progressListener = dl.QueryInterface(Ci.nsIWebProgressListener);
          persist.saveURI(dl.source, null, null, null, null, dl.targetFile);

          // This replaces all of the above, but doesn't behave as well.
          //window.location = thisDownload.getAttribute('uri');
          //gDownloadViewController.doCommand('cmd_removeFromList', thisDownload);
          break;

        default:
          break;
        }
      }
    } catch (e) {}
  }
}

////////////////////////////////////////////////////////////////////////////////
// Same as showDownload() in Firefox's downloads.js, except folder is opened
// even if file doesn't exist.
function dmtShowDownload(aItem)
{
  var fileExists = false;
  var dirExists = false;

  // Get file info about the selected download.
  try {
    var fileName = null;
    if (aItem)
    {
      // Show folder of the specified item, first need to get the download.
      var aDownload = aItem;
      while (aDownload.nodeName != "richlistitem" ||
             aDownload.getAttribute("type") != "download")
        aDownload = aDownload.parentNode;

      fileName = aDownload.getAttribute("file");
    }
    else
    {
      // No parameter so show folder for the selected download.
      fileName = gDownloadsView.selectedItem.getAttribute("file");
    }
    var fileObj = getLocalFileFromNativePathOrUrl(fileName);
    var parent = fileObj.parent.QueryInterface(Components.interfaces.nsILocalFile);
    fileExists = fileObj.exists();
    if (parent)
      dirExists = parent.exists();
    else
      dirExists = false;
  } catch (e) {
    fileExists = false
    dirExists = false;
  }

  if (fileExists)
  {
    // The file exists, use the builtin reveal() function.
    try {
      fileObj.reveal();
      return;
    } catch (e) {
      // If reveal fails then drop down to see if dirExists.
    }
  }

  if (dirExists)
  {
    // The file doesn't exist but the directory does, try using the parent.
    try {
      // "Double click" the parent directory to show where the file should be
      // This was copied from the built in Firefox download manager.
      window.parent.launch();
    } catch (e) {
      // If launch also fails (probably because it's not implemented), let the
      // OS handler try to open the parent
      openExternal(parent);
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
// Deletes the file associated with a specified download. Only done if file
// exists and user confirms the delete. Can also remove the download from the
// list, if the user selected that.
function dmtDeleteFile(aDownload, showCancel, noPrompt)
{
  var doDelete = 2;

  try
  {
    // Get a file object from the file name.
    var fileName = aDownload.getAttribute("file");
    var filePath = aDownload.getAttribute("path");
    var fileObj = getLocalFileFromNativePathOrUrl(fileName);

    if (noPrompt)
    {
      // This is only true for files after yes to all is selected.
      doDelete = 0;
    }
    else
    {
      var selected = {radio: dmtDeleteConfirmRadio, button: 3};

      // Prompt the user to make sure they want to delete each file.
      window.openDialog("chrome://downloadmgr/content/deleteConfirm.xul", "dlg",
                        "chrome,dialog,modal,resizable=no,centerscreen",
                        filePath, showCancel, selected).focus();

      dmtDeleteConfirmRadio = selected.radio;
      doDelete = selected.button;
    }

    if ((doDelete == 0) || (doDelete == 1))
    {
      if ((dmtDeleteConfirmRadio == 0) || (dmtDeleteConfirmRadio == 1))
      {
        // Make sure file exists before trying to delete it.
        if (fileObj.exists())
        {
          // Selected yes or yes to all and radio is for a delete option.
          // Force file to writeable and then delete. Note 0666 is octal.
          fileObj.permissions |= 0666;
          fileObj.remove(false);
          dmtUpdateDownloadButtons(aDownload);
        }
      }
      if ((dmtDeleteConfirmRadio == 1) || (dmtDeleteConfirmRadio == 2))
      {
        // Selected yes or yes to all and radio is for a remove option.
        // Remove the download from the list.
        gDownloadViewController.doCommand('cmd_removeFromList', aDownload);
      }
    }
  } catch (e) {
    doDelete = 2;
  }

  return doDelete;
}

////////////////////////////////////////////////////////////////////////////////
// Loops through all the selected downloads and deletes the associated file.
// Only done if files exist and user confirms the delete. Can also remove the
// downloads from the list, if the user selected that.
function dmtDeleteFiles(aItem)
{
  var doDelete = 0;
  var fileWasDeleted = false;
  var deleteAllSelected = false;
  var defaultConfirmBoth = false;
  var deleteKeyRemoves = false;

  try {
    defaultConfirmBoth = dmtPreferencesService.getBoolPref(
                                  "extensions.downloadmgr.defaultConfirmBoth");
    deleteKeyRemoves = dmtPreferencesService.getBoolPref(
                                  "extensions.downloadmgr.deleteKeyRemoves");
  } catch (e) {}

  if (deleteKeyRemoves)
  {
    performCommand('cmd_removeFromList');
    return;
  }

  if (defaultConfirmBoth)
    dmtDeleteConfirmRadio = 1;
  else
    dmtDeleteConfirmRadio = 0;

  if (aItem)
  {
    // Delete the file for the specified item, first need to get the download
    var aDownload = aItem;
    while (aDownload.nodeName != "richlistitem" ||
           aDownload.getAttribute("type") != "download")
      aDownload = aDownload.parentNode;

    doDelete = dmtDeleteFile(aDownload, false, false);

    if (doDelete == 1)       // Selected yes.
      fileWasDeleted = true;
  }
  else
  {
    var fileCount = gDownloadsView.selectedItems.length;
    var deletelist = [];

    // Need to copy the selected list because the delete process can change the
    // original list, if the remove from list option is selected.
    deletelist[fileCount - 1] = null;
    for (let i = 0; i < fileCount; i++)
      deletelist[i] = gDownloadsView.selectedItems[i];

    // Do the file delete for each download in the list of selected downloads.
    for (let i = 0; i < fileCount; i++)
    {
      doDelete = dmtDeleteFile(deletelist[i], fileCount > 1, deleteAllSelected);

      if (doDelete == 0)          // Selected yes to all.
      {
        fileWasDeleted = true;
        deleteAllSelected = true;
      }
      else if (doDelete == 1)     // Selected yes.
        fileWasDeleted = true;
      else if (doDelete == 3)     // Selected cancel.
        break;
    }
  }

  // Update button greyouts in toolbar, if a file was deleted.
  if (fileWasDeleted)
  {
    var dlState = 0;

    // Get parameter for dmtUpdateToolInfoBar().
    if ((gDownloadsView.selectedCount > 0) && gDownloadsView.selectedItem)
    {
      dlState = parseInt(gDownloadsView.selectedItem.getAttribute("state"));
    }
    dmtUpdateToolInfoBar(dlState);
  }
}

////////////////////////////////////////////////////////////////////////////////
// The user can select optional actions to take when a download is added. This
// performs those selected actions.
function dmtDownloadAddedAction(thisDownload)
{
  var limitCount = false;
  var countLimit = 0;
  var limitAge = false;
  var ageLimit = 0;

  // Get the preferences for the optional actions.
  try {
    limitCount = dmtPreferencesService.getBoolPref(
                                  "extensions.downloadmgr.limitDownloadCount");
    countLimit = dmtPreferencesService.getIntPref(
                                  "extensions.downloadmgr.downloadCountLimit");
    limitAge = dmtPreferencesService.getBoolPref(
                                  "extensions.downloadmgr.limitDownloadAge");
    ageLimit = dmtPreferencesService.getIntPref(
                                  "extensions.downloadmgr.downloadAgeLimit");
  } catch (e) {}

  // Remove downloads if count is larger than user defined limit.
  if (limitCount && (countLimit > 0) && (gDownloadsView.itemCount > countLimit))
  {
    // Remove downloads from the end of the list.
    for (let i = gDownloadsView.itemCount - 1; i >= countLimit; i--)
    {
      var thisDownload = gDownloadsView.getItemAtIndex(i);

      // Note that doCommand will not remove if download is not "removable".
      gDownloadViewController.doCommand('cmd_removeFromList', thisDownload);
    }
  }

  // Remove downloads if age is older than user defined limit.
  if (limitAge && (ageLimit > 0))
  {
    var currentTime = Date.now();
    // Remove downloads from the end of the list.
    for (let i = gDownloadsView.itemCount - 1; i >= 0; i--)
    {
      var thisDownload = gDownloadsView.getItemAtIndex(i);
      var downloadTime = parseInt(thisDownload.getAttribute("endTime"));

      // Times are in millisec but ageLimit is in days.
      if ((currentTime - downloadTime) > ageLimit * 86400000)
      {
        // Note that doCommand will not remove if download is not "removable".
        gDownloadViewController.doCommand('cmd_removeFromList', thisDownload);
      }
    }
  }
}

////////////////////////////////////////////////////////////////////////////////
// The user can select optional actions to take when a download completes. This
// performs those selected actions.
function dmtDownloadCompleteAction(thisDownload)
{
  var autoLaunch = false;
  var autoRemove = false;

  // Get the preferences for the optional actions.
  try {
    autoLaunch = dmtPreferencesService.getBoolPref(
                                  "extensions.downloadmgr.autoLaunch");
    autoRemove = dmtPreferencesService.getBoolPref(
                                  "extensions.downloadmgr.autoRemove");
  } catch (e) {}

  // Perform the actions if the preferences are set. Important that the remove
  // action is last.
  if (autoLaunch)
  {
    // User selected the option to launch downloads that have completed.
    gDownloadViewController.doCommand('cmd_open', thisDownload);
  }
  if (autoRemove)
  {
    // User selected the option to remove downloads that have completed.
    gDownloadViewController.doCommand('cmd_removeFromList', thisDownload);
  }
}

////////////////////////////////////////////////////////////////////////////////
// Sometimes rescheduling at the end of the event queue isn't enough. This does
// it a second time.
function dmtUpdateDownloadButtonsById(downloadId)
{
  var thisDownload = getDownload(downloadId);
  if (thisDownload != null)
    dmtUpdateDownloadButtons(thisDownload);
}

////////////////////////////////////////////////////////////////////////////////
// Sometimes rescheduling at the end of the event queue isn't enough. This does
// it a second time.
function dmtUpdateDownloadIconClassesById(downloadId)
{
  var thisDownload = getDownload(downloadId);
  if (thisDownload != null)
    dmtUpdateDownloadIconClasses(thisDownload);
}

////////////////////////////////////////////////////////////////////////////////
// The dmtDownloadProgressListener class is used to update the toolbar and
// infobar for downloads in progress.
// This class implements the nsIDownloadProgressListener interface.
function dmtDownloadProgressListener() {}

dmtDownloadProgressListener.prototype = {
  QueryInterface: XPCOMUtils.generateQI([Ci.nsIDownloadProgressListener]),

  onDownloadStateChange: function dlPL_onDownloadStateChange(aState, aDownload)
  {
    // One of the downloads changed state.
    var thisDownload = getDownload(aDownload.id);

    // If the changed download is the same as the selected download then the
    // toolbar needs to be updated.
    if (gDownloadsView.selectedItem == thisDownload)
    {
      // Update the entire toolbar and infobar, mainly the button greyouts.
      dmtUpdateToolInfoBar(aDownload.state);
    }

    // If the download changed into the finished state then the download buttons
    // need to be updated.
    switch (aDownload.state)
    {
    case nsIDM.DOWNLOAD_FINISHED:
      // Note that at this instant the download has not been updated to display
      // the buttons that need to be updated. This schedules the function call
      // to happen after the handlers for this event finish, which is after the
      // display has been updated.
      window.setTimeout(dmtUpdateDownloadButtons, 0, thisDownload);
      window.setTimeout(dmtDownloadCompleteAction, 0, thisDownload);
      break;

    case nsIDM.DOWNLOAD_BLOCKED_PARENTAL:
    case nsIDM.DOWNLOAD_SCANNING:
    case nsIDM.DOWNLOAD_DIRTY:
    case nsIDM.DOWNLOAD_BLOCKED_POLICY:
      // No buttons displayed in these states.
      break;

    case nsIDM.DOWNLOAD_FAILED:
    case nsIDM.DOWNLOAD_CANCELED:
      // Apply the button classes to all the buttons. The delay reschedules
      // the call to the end on the event queue, after the handlers have been
      // notified and have updated the download to display the buttons that
      // need updating.
      // This is separate from below because the added Remove From List button
      // needs to be hidden or not.
      window.setTimeout(dmtUpdateDownloadButtons, 0, thisDownload);
      break;

    case nsIDM.DOWNLOAD_DOWNLOADING:
    case nsIDM.DOWNLOAD_PAUSED:
    case nsIDM.DOWNLOAD_NOTSTARTED:
    case nsIDM.DOWNLOAD_QUEUED:
    default:
      // Apply the button classes to all the buttons. The delay reschedules
      // the call to the end on the event queue, after the handlers have been
      // notified and have updated the download to display the buttons that
      // need updating.
      if (thisDownload == null)
        window.setTimeout(dmtUpdateDownloadButtonsById, 0, aDownload.id);
      else
      {
        window.setTimeout(dmtUpdateDownloadButtons, 0, thisDownload);
        // Starting in 4.0, need an extra event to keep buttons classes set.
        //window.setTimeout(dmtUpdateDownloadButtons, 100, thisDownload);
      }
//        window.setTimeout(dmtUpdateDownloadIconClassesById, 0, aDownload.id);
//      else
//        window.setTimeout(dmtUpdateDownloadIconClasses, 0, thisDownload);

      if (aDownload.state == nsIDM.DOWNLOAD_QUEUED)
      {
        // Note that at this instant the download has not been updated to display
        // the buttons that need to be updated. This schedules the function call
        // to happen after the handlers for this event finish, which is after the
        // display has been updated.
        window.setTimeout(dmtDownloadAddedAction, 0, thisDownload);
      }
      break;
    }
  },

  onProgressChange : function(aWebProgress,      aRequest,
                              aCurSelfProgress,  aMaxSelfProgress,
                              aCurTotalProgress, aMaxTotalProgress,
                              aDownload)
  {
    var thisDownload = getDownload(aDownload.id);

    if (thisDownload != null)
    {
      // Starting in 4.0, need an extra event to keep buttons classes set.
      dmtUpdateDownloadIconClasses(thisDownload);
      //dmtUpdateDownloadButtons(thisDownload);
      //window.setTimeout(dmtUpdateDownloadButtons, 0, thisDownload);
    }
  },

  // These functions are called as part of the nsIDownloadProgressListener
  // interface, but are unused by DMT.
  onStateChange : function(prog, req, flags, status, dl) {
  },
  onSecurityChange : function(prog, req, state, dl) {
  },
};

////////////////////////////////////////////////////////////////////////////////
// Class to handle notifications from nsIObserverService
var dmtDownloadmgrOverlayObserver = {
  observe: function (aSubject, aTopic, aState)
  {
    switch (aTopic) {
    case "download-manager-remove-download":
      // This handles the events generated when downloads have been removed from
      // the list. May need to update the tool and info bar if the selected
      // download has been removed. Note that setTimeout is needed because the
      // list hasn't been updated yet but will be when the timeout event is
      // handled.
      window.setTimeout(dmtOnDownloadSelect, 0, 0);
      break;

    case "download-manager-ui-done":
      // This event is generated when the download view has finished being
      // filled with downloads.

      // Need to remove old downloads, if user has requested such.
      dmtDownloadAddedAction(null);

      // Need to update those buttons that need to be hidden.
      dmtUpdateDownloadViewButtons();
      break;
    }
  }
};

////////////////////////////////////////////////////////////////////////////////
// Called when a shortcut key is used to close the download manager. The close
// has to happen in browserOverlay.js so a notify is used to transfer control.
function dmtCloseDownloadMgr()
{
  dmtObserverService.notifyObservers(null, "DMT-ui-close", null);

  // If DMT is open in a separate window and that window is the only one open
  // (all the other windows have been closed) then the observer service is
  // shutdown and the above notify won't be sent. This closes that window in
  // that case.
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].
                      getService();
  var windowManagerInterface = windowManager.QueryInterface(
                                       Components.interfaces.nsIWindowMediator);
  var windowObj = windowManagerInterface.getMostRecentWindow("Download:Manager");

  if (windowObj)
    windowObj.close();
}

////////////////////////////////////////////////////////////////////////////////
// Called when a shortcut key is used to close the download manager. The close
// has to happen in browserOverlay.js so a notify is used to transfer control.
function dmtEscapeDownloadMgr()
{
  dmtObserverService.notifyObservers(null, "DMT-ui-escape", null);

  // If DMT is open in a separate window and that window is the only one open
  // (all the other windows have been closed) then the observer service is
  // shutdown and the above notify won't be sent. This closes that window in
  // that case.
  var windowManager = Components.classes['@mozilla.org/appshell/window-mediator;1'].
                      getService();
  var windowManagerInterface = windowManager.QueryInterface(
                                       Components.interfaces.nsIWindowMediator);
  var windowObj = windowManagerInterface.getMostRecentWindow("Download:Manager");

  if (windowObj)
    windowObj.close();
}

////////////////////////////////////////////////////////////////////////////////
// This function is overwritting the one in downloads.js so tab and sidebar will
// also be closed when it is called.
function gCloseDownloadManager()
{
  dmtCloseDownloadMgr();
}

////////////////////////////////////////////////////////////////////////////////
// This overrides the function of the same name in downloads.js and allows the
// window title or tab title to be named differently from the default. It also
// provides a way for the sidebar title to be changed.
//function onUpdateProgress()
//{
//  let numActiveDownloads = gDownloadManager.activeDownloadCount;
//
//  // Use the default title and reset "last" values if there's no downloads
//  if (numActiveDownloads == 0) {
//    document.title = document.documentElement.getAttribute("statictitle");
//    gLastComputedMean = -1;
//    gLastActiveDownloads = 0;
//  }
//  else
//  {
//    document.title = document.documentElement.getAttribute("statictitle") + " = " + numActiveDownloads;
//  }
//
//  if (dmtThisIsASidebar)
//  {
//    // Change title of Download Manager in a side bar. There could be multiple
//    // sidebars so try them all until a match is found. All sidebars have the same
//    // fairly unique class that we can search on.
//    var sidebarElements = window.top.document.getElementsByClassName(
//                                                  "chromeclass-extrachrome");
//    for (var i = 0; i < sidebarElements.length; i++)
//    {
//      var sidebarWindows = sidebarElements[i].getElementsByTagName("browser");
//      for (var j = 0; j < sidebarWindows.length; j++)
//      {
//        if (sidebarWindows[j].contentWindow.location.href ==
//            "chrome://mozapps/content/downloads/downloads.xul")
//        {
//          var sidebarHeaders = sidebarElements[i].getElementsByTagName("sidebarheader");
//          var sidebarLabel = sidebarHeaders[j].getElementsByTagName("label");
//          sidebarLabel[0].value = document.title;
//        }
//      }
//    }
//  }
//}
